// module clientsocket.d
// Copyright Dr.E.Huckert 02-2018
// This is a functional subset of my class CLIENTSOCKET

// versions; // end CLIENTSOCKET::open()
//   2018-02-10 first working version for prog.lang. D
//   2018-02-20: send() uses now single TCP-send call: huge perf.improvment

// V1.X27

module clientsocket;

import std.socket;
import msgsocket;
import std.string;
import core.thread;
import std.conv;
import d_trace;

class CLIENTSOCKET
{
private:
  std.socket.Socket  tcpSnd;
  char [] inetaddr;
  long    timeout_ms;
  std.socket.socket_t handle;
  Socket[] reads;  // for select()/testDataAvailable()
  int     protocol;
  int     traceLevel;
  D_TRACE pTrace;
  ulong   nettoRx;
  ulong   nettoTx;
  void    log(const char *msg);
  void    log(const char *msg, const char *sarg);
  void    log(const char *msg, int intarg);
  void    log(const char *msg, long longarg);
  void    logError(const char *msg);
public:
  // constructor
  this()
  {
    this.tcpSnd = null;
    this.protocol = 0;
    this.pTrace = null;
    this.timeout_ms = 100;
  };
  // destructor
  ~this()
  {
    if (tcpSnd !is null)
      delete tcpSnd;
  };
  // ------------------------------------------------
  //
  // May return -1 if socket is not open!
  std.socket.socket_t get_handle()
  { 
    if (this.tcpSnd !is null)
      return this.tcpSnd.handle;
    else
      return cast(std.socket.socket_t)-1;
  };
  // -----------------------------------------------
  // Tests if data are available on the socket (for read only)
  bool testDataAvailable(uint ms)
  {
    bool ret = false;
    TimeVal timeout;
    //
    timeout.seconds      = 0;
    timeout.microseconds = ms * 1000; 
    if (this.reads.length > 0)
    {
      auto sSet = new SocketSet(1);
      sSet.add(this.tcpSnd);
      Socket.select(sSet, null, null, &timeout);
      if (sSet.isSet(reads[0]))
        ret = true;
      delete sSet;
    }
    //
    return ret;
  };   // end CLIENTSOCKET::testDataAvailable()
  //
  // -----------------------------------------------
  int open(const string hostname, int portnumber)
  {
    int ret = 0;
    //
    try
    {
      this.tcpSnd = new TcpSocket(AddressFamily.INET);
      //ulong [1] opts;
      //opts[0] = 1L;
      this.tcpSnd.setOption(SocketOptionLevel.SOCKET,
                            SocketOption.TCP_NODELAY, true);
      this.tcpSnd.connect(new InternetAddress(hostname, 
                          cast(ushort)portnumber));
      this.reads ~= this.tcpSnd;
    }
    catch(SocketException e)
    {
      ret = -1;
    }
    if (this.pTrace !is null)
      this.pTrace.trace("CL(D)::open() ret=", ret);
    return ret;
  };   // end CLIENTSOCKET::open()
  //
  // -----------------------------------------------------
  void  close()
  {
    if (this.pTrace !is null)
      this.pTrace.trace("CL(D)::close() ");
    if (tcpSnd !is null)
      tcpSnd.close();
    return;
  };
  // ------------------------------------------------------
  int getProtocol()
  { return this.protocol; };
  void setProtocol(int prot)
  { this.protocol = prot; };
  //
  // --------------------------------------------------------
  void milli_sleep(int ms)
  {
    Thread.sleep(dur!("msecs")(ms));
  };
  // --------------------------------------------------------
  void   setLogging(D_TRACE pt, int level)
  {
    this.traceLevel = level;
    this.pTrace     = pt;
  };
  // --------------------------------------------------------
  void   set_timeout(int secs)
  {
    this.timeout_ms = secs * 1000;
  };
  // --------------------------------------------------------
  void   set_timeout_ms(int ms)
  {
    this.timeout_ms = ms;
  }
    // --------------------------------------------------------
  int    getError()
  {
    int ret = 0;
    return ret;
  };
  //
  // -----------------------------------------------
  // Returns: the number of bytes sent
  // Returns: < 0 upon error
  int send(MSGSOCKET pmsg, int wait_flag)
  {
    int ret = 0;
    ubyte[] sndBuf;
    //
    if (this.pTrace !is null)
      this.pTrace.trace("CL(D)::send() length content=",
                         pmsg.get_content_length());
    if (pmsg is null)
    {
      ret = -1;
      goto zurueck;
    }
    //
    try
    {
      uint noBytes = 0;
      char [6] auxStr;
      ubyte [] auxBuf;
      //
      noBytes = 1 + 6 + cast(uint)(pmsg.get_content_length());
      //
      // insert 6 bytes length first
      sndBuf.length = 0;
      sformat(auxStr, "%06u", noBytes - 6);
      auxBuf  = cast(ubyte[])auxStr;
      //
      // append the length
      sndBuf = sndBuf ~ auxBuf;
      sndBuf.length = 6;
      //
      sndBuf = sndBuf ~ cast(ubyte)pmsg.get_operation();
      sndBuf.length = 7;
      //
      sndBuf = sndBuf ~ pmsg.get_content();
      sndBuf.length = noBytes;
      //
      if (this.pTrace !is null)
        this.pTrace.trace("CL(D)::send() length package=",
                         sndBuf.length);
      ret = cast(int)(this.tcpSnd.send(sndBuf));
      if (this.pTrace !is null)
        this.pTrace.trace("CL(D)::send TX ret=", ret);
      this.setNettoTx(this.getNettoTx() + ret);
      if ((ret == Socket.ERROR) || (ret == 0))
      {
        if (this.pTrace !is null)
          this.pTrace.trace("CL(D)::send() socket ERROR :", 
                            std.socket.lastSocketError());
        ret = -2;
        goto zurueck;
      }
      if (ret != noBytes)
      {
        if (this.pTrace !is null)
          this.pTrace.trace("CL(D)::send() ERROR: less bytes sent than expected: noBytes=", 
                            noBytes);
        ret = -3;
        goto zurueck;
      }
    }
    catch (SocketException e)
    {
      if (this.pTrace !is null)
          this.pTrace.trace("CL(D)::send() socket EXCEPTION :", 
                            std.socket.lastSocketError());
      ret = -5;
    }
    //
    zurueck:
    delete sndBuf;
    if (this.pTrace !is null)
      this.pTrace.trace("CL(D)::send() ret=", ret);
    return ret;
  };   // end CLIENTSOCKET:.send()
  //
  // -----------------------------------------------
  // Receive a single portion of data (MSGSOCKET package) via TCP
  // Returns in any case a MSGSOCKET instance if OK
  // This returned instance must be freed/deleted externally
  // If an error occurs the operation member is set to "e"
  MSGSOCKET recv()
  {
    ubyte [2055] rxBuf;
    uint     totalNoBytes = 0;
    MSGSOCKET msgSock = new MSGSOCKET(' ', "");  // the returned instance
    int      noBytes;
    int      reason = 0;
    char     op;
    char [6] strLength;
    int      binLength;
    int      state = 0;
    uint     loopCount = 0;
    ubyte [] finalRxBuf;
    //
    if (this.pTrace !is null)
      this.pTrace.trace("CL(D) CLIENTSOCKET::recv() start");
    try
    {
      // we must receive in a loop - a logical package may be divided
      // in several TCP packages
      while (loopCount < 5)
      {
        int noMs = 500;
        while (noMs > 0)
        {
          if (! this.testDataAvailable(2)) 
          {
            this.milli_sleep(2);
            noMs -= 2;
          }
          else
            break;
        }   // end while noMs ...
        //
        // data available here
        noBytes      = cast(int)(this.tcpSnd.receive(rxBuf));
        if (this.pTrace !is null)
            this.pTrace.trace("CL(D)::recv() RX  noBytes=", noBytes);
         //
         if (loopCount == 0)
         {
           // decode the ASCII length string
           strLength = cast(char [])rxBuf[0..6];
           binLength = to!int(cast(string)strLength);
           if (this.pTrace !is null)
             this.pTrace.trace("CL(D)::recv() binLength=", 
                                binLength);
           //
           op = cast(char)rxBuf[6];
           if (this.pTrace !is null)
              this.pTrace.trace("CL(D)::recv() op=", op);
           msgSock.set_operation(op);
         }
         //
         uint from = 0;
         if (loopCount == 0) 
           from = 7;
         totalNoBytes += noBytes;
         finalRxBuf = finalRxBuf ~  rxBuf[from .. noBytes];
         if (this.pTrace !is null)
            this.pTrace.trace("CL(D)::recv() totalNoBytes=", totalNoBytes);
         if (totalNoBytes >= binLength)
         {
           if (this.pTrace !is null)
            this.pTrace.trace("CL(D)::recv() package complete!");
           reason = 3;
           break;
         }
         loopCount++;
      }   // end while (loopCount < 5)  
    }   // end try
    catch(SocketException e)
    {
      msgSock.set_operation('e');
      msgSock.set_content(cast(ubyte [])"EXCEPTION");
      reason = -2;
      goto zurueck;
    }
    //
    msgSock.set_content_length(totalNoBytes - 7);
    msgSock.set_content(finalRxBuf);
    if (this.pTrace !is null)
        this.pTrace.trace("CL(D)::recv() content length=", 
                           msgSock.get_content_length());
    //
    zurueck:
    if (this.pTrace !is null)
    {
      this.pTrace.trace("CL(D)::recv() reason=", reason);
      this.pTrace.trace("CL(D)::recv() state=", state);
    }
    return msgSock;
  };   // end MSGSOCKET recv()
  //
  // -----------------------------------------------
  // Returns: < 0 upon error
  int    basic_receive(char *buf, int lbuf, int get_all)
  {
    int ret = 0;
    return ret;
  };
  // -----------------------------------------------
  // Returns: < 0 upon error
  int    receive_line(char *buf, int lbuf)
  {
    int ret = 0;
    return ret;
  };
  //
  int    get_address(char *pretBuf, int lRetBuf)
  {
    int ret = 0;
    return ret;
  };
  int    getAddress(char *pretBuf, int lRetBuf)
  {
    int ret = 0;
    return ret;
  };
  ulong getNettoRx()
  { return this.nettoRx; };
  void setNettoRx(ulong nrx)
  { this.nettoRx = nrx; };
  ulong getNettoTx()
  { return this.nettoTx; };
  void setNettoTx(ulong ntx)
  { this.nettoTx = ntx; };
};
