// module  dTrace.d
// by E.Huckert 03-2016
// A (simplified) class for logging via UDP (see also hu_trace.cpp)
// Tested with dmd V2.070 

// V 1.X27

module d_trace;

import std.stdio;
import std.string;
import std.socket;
import std.conv;

// --------------------------------------------------------------------------
class D_TRACE
{
  private:
    string hostName;
    ushort port;
    UdpSocket udps;
    int socketError;
  public:
    // ----------------------------------------------------------------
    // constructor
    this(string host, ushort portNumber)
    {
      this.socketError = 0;
      this.hostName    = host;
      this.port        = portNumber;
      try
      {
        const char [] cHostName = this.hostName.dup;  // must be!
        Address [] addr = getAddress(cHostName, this.port);
        if (addr[0] !is null)
        {
          this.udps = new UdpSocket(AddressFamily.INET);
          this.udps.connect(new InternetAddress(this.hostName, this.port));
        }
        //writeln("D_TRACE constructor OK");
      }
      catch (SocketException e)
      {
        this.socketError = -1;
        writeln("module d_trace: EXCEPTION socket");
      }
    };   // end this()
    //
    // -----------------------------------------------------------------------
    // destructor
    ~this()
    {
      if (this.udps !is null)
        this.udps.close();
    }
    //
    // ------------------------------------------------------------------------
    // Returns < 0 upon error
    int trace(string message)
    { 
      int ret = 0;
      message ~= "\n";
      if (this.udps !is null)
      {
        ret = cast(int)(this.udps.send(message));
        if (ret == Socket.ERROR)
        ret = -2;
        else 
          ret =cast(int)message.length;
      }
      else
        ret = -1;
      return ret;
    };   // end trace()
    //
    // ------------------------------------------------------------------------
    int trace(string message1,
              string message2)
    { 
      int ret = 0;
      message1 = message1 ~ message2;
      message1 ~= "\n";
      ret = cast(int)(this.udps.send(message1));
      //
      return ret;
    };   // end trace()
    //
    // ------------------------------------------------------------------------
    int trace(string message1,
              const char [] message2)
    { 
      int ret = 0;
      ret = this.trace(message1, cast(string)message2);
      return ret;
    };   // end trace()
    //
    // -----------------------------------------------------------------------
    int trace(string message, int p1)
    { 
      int ret = 0;
      string msg2 = message;
      msg2 = msg2 ~ to!string(p1);
      ret = this.trace(msg2);
      //
      return ret;
    };   // end trace()
    //
    // -----------------------------------------------------------------------
    int trace(string message, ulong p1)
    { 
      int ret = 0;
      string msg2 = message;
      msg2 = msg2 ~ to!string(p1);
      ret = this.trace(msg2);
      //
      return ret;
    };   // end trace()
    //
    // -----------------------------------------------------------------------
    int getError()
    {
      return this.socketError;
    };
  } //end class
