// module srvsocket.d
// Copyright Dr.E.Huckert 02-2018
// This is a functional subset of my class SRVSOCKET

// versions;
//   2018-02-10 first working version for prog.lang. D

// V1.X27

module srvsocket;

import std.socket;
import msgsocket;
import d_trace;
import std.stdio;
import std.string;
import std.conv;   // needed for to!int()
import core.thread;

uint LENGTH_LENGTH = 6;  // # length bytes before netto content

class SRVSOCKET
{
private:
  Socket  listenSock;
  Socket  rwSock;
  Socket[] reads;  // for select()/testDataAvailable()
  int     protocol;    // 0 = TCP
  long    timeout_ms;
  int     sleep_unit;
  int     partial;
  ushort  port;
  ulong   nettoTx;  // gesendete Nettodaten (Nutzdaten)
  ulong   nettoRx;  // empfangene Nettodaten (Nutzdaten)
  int     traceLevel;
  D_TRACE tracer;

public:
  this()
  {
    this.basic_init(0);
  };
  this(int udp)
  {
  };
  this(Socket listenSock, Socket rwSock)
  {
  };
  // Destructor
  ~this()
  {
    // cose both sockets (if open)
    // deletes the socket instances
    this.close();
  };
  //
  // ---------------------------------------------------
  void basic_init(int proto)
  {
    this.timeout_ms      = 1000L;  // milli seconds = 1 seconds
    this.listenSock      = null; // INVALID_SOCKET;
    this.rwSock          = null; // INVALID_SOCKET;
    this.protocol        = proto;  // TCP oder UDP
    this.tracer          = null;
    this.traceLevel      = 0;
    this.partial         = 0;
    this.port            = 0;
    this.protocol        = 0;   // ASCII/ANSI oder binaer
    this.nettoRx         = 0L;  // empf. Nettodaten
    this.nettoTx         = 0L;  // ges. Nettodaten
  };
  //
  // -------------------------------------------------------
  void milli_sleep(int ms)
  {
    Thread.sleep(dur!("msecs")(ms));
  };
  //
  // -------------------------------------------------------
  void setLogging(D_TRACE pt)
  {
    this.tracer = pt;
    if (this.tracer !is null)
      this.tracer.trace("SR(D)::setLogging()");
    return;
  };
  // -------------------------------------------------------
  void setLogging(D_TRACE pt, int level)
  {
    this.tracer = pt;
    this.traceLevel = level;
    if (this.tracer !is null)
      this.tracer.trace("SR(D)::setLogging() level=", level);
    return;
  }
  int  getProtocol()
  {  return this.protocol; };
  void setProtocol(int protNo)
  {  this.protocol = protNo; };
  int  getPort()
  {  return this.port; };
  ulong getNettoRx()
  { return this.nettoRx; };
  void setNettoRx(ulong nrx)
  { this.nettoRx = nrx; };
  ulong getNettoTx()
  { return this.nettoTx; };
  void setNettoTx(ulong ntx)
  { this.nettoTx = ntx; };
  Socket getListenSocket() 
  { return this.listenSock; };
  Socket getrwSock() 
  { return this.rwSock; };
  // -----------------------------------------------
  // Tests if data are available on the socket (for read only)
  bool testDataAvailable(uint ms)
  {
    bool ret = false;
    TimeVal timeout;
    timeout.seconds      = 0;
    timeout.microseconds = ms * 1000;
    if (this.reads.length > 0)
    {
      auto sSet = new SocketSet(1);
      sSet.add(this.rwSock);
      Socket.select(sSet, null, null, &timeout);
      if (sSet.isSet(reads[0]))
        ret = true;
      delete sSet;
    }
    if (this.tracer !is null)
         this.tracer.trace("SR(D)::testDataAvailable() ret=", ret);
    return ret;
  };   //end SRVSOCKET::testDataAvailable()
  //
  // ------------------------------------------------------
  // Open a listen socket
  // Return >= 0 if OK
  int  open(ushort portNo)
  {
    int ret = 0;
    //
    if (this.tracer !is null)
      this.tracer.trace("SR(D)::open() port=", portNo);
    this.port = portNo;
    try
    {
      if (this.protocol == 0)    // TCP here
      {
        this.listenSock = new TcpSocket();
        (this.listenSock).bind(new InternetAddress(this.port)); 
      }
      else
      {
        if (this.tracer !is null)
          this.tracer.trace("SR(D)::open() protocol not supported");
        ret = -2;
      }
    }
    catch (SocketException e)
    {
      writeln("SR(D)::open(): socket EXCEPTION: ",
               std.socket.lastSocketError());
      if (this.tracer !is null)
        this.tracer.trace("SR(D)::open() socket EXCEPTION: ",
                           std.socket.lastSocketError());
      ret = -1;
    }
    if (this.tracer !is null)
         this.tracer.trace("SR(D)::open() ret=", ret);
    return ret;
  };   // end SRVSOCKET::open()
  //
  // -------------------------------------------------------
  // Accept an incoming client request
  // Result: the client socket (here called this.rwSock)
  // Returns < 0 upon error
  int  accept()
  {
    int ret = 0;
    try
    {
      this.rwSock = (this.listenSock).accept();
      this.reads ~= this.rwSock;
      //ulong [1] opts;
      //opts[0] = 1;
      this.rwSock.setOption(SocketOptionLevel.SOCKET,
                            SocketOption.TCP_NODELAY, true);
      ret = 1;
    }
    catch (SocketException e)
    {
      writeln("SRVSOCKET EXCEPTION: socket (accept)");
      ret = -1;
    }
    if (this.tracer !is null)
         this.tracer.trace("SR(D)::accept() ret=", ret);
    return ret;
  };
  // ------------------------------------------------------
  // close both sockets, delete the instances
  int  close()
  {
    int ret = 0;
    if (this.listenSock !is null)
    {
      (this.listenSock).close(); 
      delete this.listenSock;
      this.listenSock = null;
    }
    if (this.rwSock !is null)
    {
      (this.rwSock).close();
      delete this.rwSock;
      this.rwSock = null;
    }
    if (this.tracer !is null)
      this.tracer.trace("SR(D)::close() ret=", ret);
    return ret;
  };
  // ------------------------------------------------------
  // close only the RW socket
  int  close_rwSocket()
  {
    int ret = 0;
    if (! (this.rwSock is null))
      (this.rwSock).close(); 
    this.rwSock = null;  
    if (this.tracer !is null)
        this.tracer.trace("SR(D)::close_rwsocket() ret=", ret);    
    return ret;
  };
  //
  // -----------------------------------------------
  // returns < 0 upon error
  int  send(MSGSOCKET msgSock, int wait_flag)
  {
    int ret = 0;
    try
    {
       ubyte [] sndBuf;
       char [6] cBuf;
       int nb;
       // 19.02.2018: send in a single call
       sndBuf.length = 0;
       nb = cast(int)((msgSock.get_content()).length + 6 + 1);
       if ((this.tracer !is null) && (this.traceLevel > 1))
         this.tracer.trace("SR(D)::send() will send nb bytes=", nb);
       sformat(cBuf, "%06u", nb);
       sndBuf = sndBuf ~ cast(ubyte[])cBuf;
       //
       // append the op.byte
       sndBuf = sndBuf ~ msgSock.get_operation();
       //
       // append the content
       sndBuf = sndBuf ~ cast(ubyte[])(msgSock.get_content());
       //
       // send the complet package now
       nb = cast(int)(this.rwSock.send(sndBuf));
       if (nb != (msgSock.get_content_length() + 6 + 1))
       {
         ret = -4;
         if (this.tracer !is null)
           this.tracer.trace("SR(D)::send() ERROR : only sent nb=", nb);
         goto zurueck;
       }
       ret = ret + nb;
       this.nettoTx += nb;
    }   // end try
    catch (SocketException e)
    {
      writeln("R(D)::send()EXCEPTION: socket (send)");
      ret = -1;
    }
    zurueck:
    if (this.tracer !is null)
      this.tracer.trace("SR(D)::send() ret=", ret);
    return ret;
  };   // end SRVSOCKET::send()
  //
  // ----------------------------------------------------
  void set_timeout(long milli_seconds)
  {
    this.timeout_ms = milli_seconds;
    return;
  };
  // -----------------------------------------------------
  // listen for an incoming client request
  // bind() must have been called
  int  listen(int noConns)
  {
    int ret = 0;
    (this.listenSock).listen(noConns);
    return ret;
  };
  // ---------------------------------------------------
  // Receive an MSGSOCKET package
  // Returns in any case a MSGSOCKET instance that must be freed externally
  MSGSOCKET recv()
  {
    
    char    opChar = ' ';
    ubyte[] rxBuf;
    int     reason = 0;
    MSGSOCKET msgSock = null;
    //
    if (this.tracer !is null)
      this.tracer.trace("SR(D)::recv(): start");
    rxBuf.length = 0;
    msgSock = new MSGSOCKET(opChar, rxBuf);
    //
    try
    {
       ubyte [] auxBuf;
       char [] cBuf;
       int nb, lBuf;
       //
       if (! testDataAvailable(500)) // wait 500 ms
       {
         msgSock.set_operation('e');
         msgSock.set_content(cast(ubyte[])"SR(D)::recv(): rec. timeout");
         reason = -1;
         goto zurueck;
       }
       // get the length (LENGTH bytes) and the op byte first
       rxBuf.length = LENGTH_LENGTH + 1;
       nb = cast(int)(this.rwSock.receiveFrom(rxBuf)); 
       if ((this.tracer !is null) && (this.traceLevel > 0))
         this.tracer.trace("SR(D)::recv(): receiveFrom(length+op) nb=", nb);
       if (nb != (LENGTH_LENGTH + 1)) 
       {
         if (this.tracer !is null)
           this.tracer.trace("SR(D)::recv() invalid package start=", nb);
         msgSock.set_operation('e');
         msgSock.set_content(cast(ubyte[])"SR(D)::recv(): invalid package start");
         reason = -2;
         goto zurueck;
       }
       char [] s = cast(char [])rxBuf[0 .. 6];
       lBuf = to!int(s);
       if ((this.tracer !is null) && (this.traceLevel > 1))
           this.tracer.trace("SR(D)::recv() lBuf=", lBuf);
       // get the operation byte
       opChar = cast(char)rxBuf[6];
       s.length = 1;
       s[0] = opChar;
       if (this.tracer !is null)
         this.tracer.trace("SR(D)::recv(): op byte=", s);
       //
       // get the content now
       if (this.tracer !is null)
         this.tracer.trace("SR(D)::recv(): receiveFrom() reading content");
       rxBuf.length = lBuf - 1;
       nb = cast(int)(this.rwSock.receiveFrom(rxBuf));
       if (this.tracer !is null)
         this.tracer.trace("SR(D)::recv(): content nb=", nb);
       if (nb != (lBuf - 1))
       {
         msgSock.set_operation('e');
         msgSock.set_content(cast(ubyte[])"SR(D)::recv() invalid content portion");
         if (this.tracer !is null)
           this.tracer.trace("SR(D)::recv()invalid content portion");
         reason = -4;
         goto zurueck;
       }
       msgSock = new MSGSOCKET(opChar, rxBuf, cast(uint)rxBuf.length);
       reason = 1;
    }
    catch (SocketException e)
    {
      msgSock.set_operation('e');
      msgSock.set_content(cast(ubyte[])"SR(D)::recv() Socket EXCEPTION");
      if (this.tracer !is null)
           this.tracer.trace("SR(D)::recv() Socket EXCEPTION");
      reason = -5;
    }
    zurueck:
    if (this.tracer !is null)
      this.tracer.trace("SR(D)::recv() reason=", reason);
    return msgSock;
  };
};   // end SRVSOCKET::recv()
